# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import pandas as pd
from pandas import DataFrame

from . import constants


def map_to_udm_grades(grades_df: pd.DataFrame) -> pd.DataFrame:
    """
    Maps a DataFrame containing Canvas grades into the Ed-Fi LMS Unified Data
    Model (UDM) format.

    Parameters
    ----------
    grades_df: DataFrame
        Pandas DataFrame containing grades

    Returns
    -------
    DataFrame
        A LMSGrades-formatted DataFrame

    Notes
    -----
    DataFrame columns are:
        SourceSystemIdentifier: A unique number or alphanumeric code assigned to a
        grade by the source system.
        LMSUserIdentifier: A unique numeric identifier assigned to the user.
        SourceSystem: The system code or name providing the section data
        LMSSectionIdentifier: A unique numeric identifier assigned to the section.
        LMSGradeIdentifier: A unique numeric identifier assigned to the grade.
        GradeType: The type of grade reported. E.g., Current, Final.
        Grade: The user's letter or numeric grade for the section.
        SourceCreateDate: Date this record was created in the LMS
        SourceLastModifiedDate: Date this record was last updated in the LMS
        CreateDate: Date this record was created in the extractor
        LastModifiedDate: Date this record was last updated in the extractor
    """

    if grades_df.empty:
        return grades_df

    assert "SourceSystemIdentifier" in grades_df.columns
    assert "LMSUserLMSSectionAssociationSourceSystemIdentifier" in grades_df.columns
    assert "final_score" in grades_df.columns
    assert "CreateDate" in grades_df.columns
    assert "LastModifiedDate" in grades_df.columns

    df: DataFrame = grades_df[
        [
            "SourceSystemIdentifier",
            "LMSUserLMSSectionAssociationSourceSystemIdentifier",
            "final_score",
            "CreateDate",
            "LastModifiedDate",
        ]
    ].copy()

    df.rename(
        columns={"final_score": "Grade"},
        inplace=True,
    )

    df["SourceSystem"] = constants.SOURCE_SYSTEM
    df["GradeType"] = "Final"
    df["SourceCreateDate"] = ""
    df["SourceLastModifiedDate"] = ""

    return df



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

from typing import Dict, List, Tuple
import logging
from canvasapi import Canvas
from canvasapi.authentication_event import AuthenticationEvent
from canvasapi.enrollment import Enrollment
import sqlalchemy
from canvasapi.course import Course
from canvasapi.section import Section
from canvasapi.user import User
from canvasapi.assignment import Assignment
from canvasapi.submission import Submission
from pandas import DataFrame

from edfi_canvas_extractor.api import (
    courses as coursesApi,
    sections as sectionsApi,
    students as studentsApi,
    assignments as assignmentsApi,
    submissions as submissionsApi,
    enrollments as enrollmentsApi,
    authentication_events as authEventsApi,
)
from edfi_canvas_extractor.mapping import (
    sections as sectionsMap,
    users as usersMap,
    assignments as assignmentsMap,
    submissions as submissionsMap,
    section_associations as section_associationsMap,
    grades as gradesMap,
    authentication_events as authEventsMap,
)

logger = logging.getLogger(__name__)


def extract_courses(
    canvas: Canvas,
    start_date: str,
    end_date: str,
    sync_db: sqlalchemy.engine.base.Engine,
) -> Tuple[List[Course], DataFrame]:
    """
    Gets all Canvas courses for the given date range, in the Ed-Fi UDM format.

    Parameters
    ----------
    canvas: Canvas
        Canvas object.
    start_date: str
        Retrieve Courses starting on or after this date.
    end_date:
        Retrieve Courses ending on or before this date.
    sync_db: sqlalchemy.engine.base.Engine
        Sync database connection.

    Returns
    -------
    Tuple[List[Course], DataFrame]
        A tuple with the list of Canvas Course objects and the udm_courses dataframe.
    """
    courses: List[Course] = coursesApi.request_courses(canvas, start_date, end_date)
    courses_df: DataFrame = coursesApi.courses_synced_as_df(courses, sync_db)

    return (courses, courses_df)


def extract_sections(
    courses: List[Course], sync_db: sqlalchemy.engine.base.Engine
) -> Tuple[List[Section], DataFrame, List[str]]:
    """
    Gets all Canvas sections, in the Ed-Fi UDM format.

    Parameters
    ----------
    courses: List[Course]
        A list of Canvas Course objects.
    sync_db: sqlalchemy.engine.base.Engine
        Sync database connection.

    Returns
    -------
    Tuple[List[Section], DataFrame, List[str]]
        A tuple with the list of Canvas Section objects, the udm_sections dataframe,
        and a list of all section ids as strings.
    """
    sections: List[Section] = sectionsApi.request_sections(courses)
    sections_df: DataFrame = sectionsApi.sections_synced_as_df(sections, sync_db)
    udm_sections_df: DataFrame = sectionsMap.map_to_udm_sections(sections_df)
    section_ids = udm_sections_df["SourceSystemIdentifier"].astype("string").tolist()
    return (sections, udm_sections_df, section_ids)


def extract_students(
    courses: List[Course], sync_db: sqlalchemy.engine.base.Engine
) -> Tuple[List[User], DataFrame]:
    """
    Gets all Canvas students, in the Ed-Fi UDM format.

    Parameters
    ----------
    courses: List[Course]
        A list of Canvas Course objects.
    sync_db: sqlalchemy.engine.base.Engine
        Sync database connection.

    Returns
    -------
    Tuple[List[User], DataFrame]
        A tuple with the list of Canvas User objects and the udm_users dataframe.
    """
    students: List[User] = studentsApi.request_students(courses)
    students_df: DataFrame = studentsApi.students_synced_as_df(students, sync_db)
    udm_students_df: DataFrame = usersMap.map_to_udm_users(students_df)

    return (students, udm_students_df)


def extract_assignments(
    courses: List[Course],
    sections_df: DataFrame,
    sync_db: sqlalchemy.engine.base.Engine,
) -> Tuple[List[Assignment], Dict[str, DataFrame]]:
    """
    Gets all Canvas assignments, in the Ed-Fi UDM format.

    Parameters
    ----------
    courses: List[Course]
        A list of Canvas Course objects.
    sections_df: DataFrame
        A DataFrame of Canvas Section objects.
    sync_db: sqlalchemy.engine.base.Engine
        Sync database connection.

    Returns
    -------
    Tuple[List[Assignment], DataFrame]
        A tuple with the list of Canvas Assignment objects and the udm_assignments dataframe.
    """
    assignments: List[Assignment] = assignmentsApi.request_assignments(courses)
    if len(list(assignments)) < 1:
        logger.info("Skipping assignments - No data returned by API")
        return ([], {})
    assignments_df: DataFrame = assignmentsApi.assignments_synced_as_df(
        assignments, sync_db
    )
    udm_assignments_dfs: Dict[str, DataFrame] = assignmentsMap.map_to_udm_assignments(
        assignments_df, sections_df
    )

    return (assignments, udm_assignments_dfs)


def extract_submissions(
    assignments: List[Assignment],
    sections: List[Section],
    sync_db: sqlalchemy.engine.base.Engine,
) -> Dict[Tuple[str, str], DataFrame]:
    """
    Gets all Canvas submissions, in the Ed-Fi UDM format.

    Parameters
    ----------
    assignments: List[Assignment]
        A list of Canvas Assignment objects.
    sections: List[Section]
        A List of Canvas Section objects.
    sync_db: sqlalchemy.engine.base.Engine
        Sync database connection.

    Returns
    -------
    Dict[Tuple[str, str], DataFrame]
        A dict with (section_id, assignment_id) as key and udm_submissions
        as value.
    """
    export: Dict[Tuple[str, str], DataFrame] = {}
    for section in sections:
        for assignment in [
            assignment
            for assignment in assignments
            if assignment.course_id == section.course_id
        ]:
            submissions: List[Submission] = submissionsApi.request_submissions(
                assignment
            )
            if len(list(submissions)) < 1:
                logger.info(
                    "Skipping submissions for assignment id %s - No data returned by API",
                    assignment.id,
                )
                continue
            submissions_df: DataFrame = submissionsApi.submissions_synced_as_df(
                submissions, sync_db
            )
            submissions_df = submissionsMap.map_to_udm_submissions(submissions_df)
            export[(str(section.id), str(assignment.id))] = submissions_df
    return export


def extract_enrollments(
    sections: List[Section], sync_db: sqlalchemy.engine.base.Engine
) -> Tuple[List[Enrollment], Dict[str, DataFrame]]:
    """
    Gets all Canvas enrollments, in the Ed-Fi UDM format.

    Parameters
    ----------
    sections: List[Section]
        A list of Canvas Section objects.
    sync_db: sqlalchemy.engine.base.Engine
        Sync database connection.

    Returns
    -------
    Dict[str, DataFrame]
        A dict with section_id as key and udm_enrollments as value.
    """
    udm_enrollments: Dict[str, DataFrame] = dict()
    enrollments: List[Enrollment] = []
    for section in sections:
        local_enrollments: List[Enrollment] = list(
            enrollmentsApi.request_enrollments_for_section(section)
        )
        if len(list(local_enrollments)) < 1:
            logger.info(
                "Skipping enrollments for section id %s - No data returned by API",
                section.id,
            )
            continue
        enrollments_df: DataFrame = enrollmentsApi.enrollments_synced_as_df(
            local_enrollments, sync_db
        )
        enrollments_df = section_associationsMap.map_to_udm_section_associations(
            enrollments_df
        )
        enrollments = enrollments + local_enrollments
        udm_enrollments[str(section.id)] = enrollments_df

    return (enrollments, udm_enrollments)


def extract_grades(
    enrollments: List[Enrollment],
    udm_enrollments: Dict[str, DataFrame],
    sections: List[Section],
) -> Dict[str, DataFrame]:
    """
    Gets all Canvas grades, in the Ed-Fi UDM format.

    Parameters
    ----------
    enrollments: List[Enrollment]
        A list of Canvas Enrollment objects.
    udm_enrollments: Dict[str, DataFrame]
        A dict of udm enrollments with section_id as the key and DataFrame as value.
    sections: List[Section]
        A list of Canvas Section objects.

    Returns
    -------
    Dict[str, DataFrame]
        A dict with section_id as key and UDM Grades DataFrame as value.
    """
    output: Dict[str, DataFrame] = {}

    for section in sections:
        current_grades: List[dict] = []
        section_id: str = str(section.id)
        if section_id not in udm_enrollments:
            logger.info(
                "Skipping enrollments for section id %s - None found", section_id
            )
            continue
        udm_enrollments_list: List[dict] = udm_enrollments[section_id].to_dict(
            "records"
        )

        for enrollment in [
            enrollment
            for enrollment in enrollments
            if enrollment.type == "StudentEnrollment"
            and enrollment.course_section_id == section.id
        ]:
            grade: dict = enrollment.grades
            current_udm_enrollment = [
                first_enrollment
                for first_enrollment in udm_enrollments_list
                if first_enrollment["SourceSystemIdentifier"] == str(enrollment.id)
            ][0]
            grade["SourceSystemIdentifier"] = f"g#{enrollment.id}"
            grade["LMSUserLMSSectionAssociationSourceSystemIdentifier"] = str(
                enrollment.id
            )
            grade["LMSSectionIdentifier"] = section_id
            grade["CreateDate"] = current_udm_enrollment["CreateDate"]
            grade["LastModifiedDate"] = current_udm_enrollment["LastModifiedDate"]
            current_grades.append(grade)

        output[section_id] = gradesMap.map_to_udm_grades(DataFrame(current_grades))

    return output


def extract_system_activities(
    users: List[User],
    start_date: str,
    end_date: str,
    sync_db: sqlalchemy.engine.base.Engine,
) -> DataFrame:
    """
    Gets all Canvas students, in the Ed-Fi UDM format.

    Parameters
    ----------
    users: List[User]
        A list of Canvas User objects.
    start_date: str
        Retrieve events occurring on or after this date.
    end_date: str
        Retrieve events occurring on or before this date.
    sync_db: sqlalchemy.engine.base.Engine
        Sync database connection.

    Returns
    -------
    DataFrame
        A Dataframe with udm_system_activities.
    """

    def _get_authentication_events():
        auth_events: List[AuthenticationEvent] = authEventsApi.request_events(
            users, start_date, end_date
        )
        if len(list(auth_events)) < 1:
            logger.info("Skipping authentication events - No data returned by API")
            return DataFrame()

        for event in auth_events:
            user_id = event.links["user"]
            event_type = event.event_type
            # according to Canvas docs, login and logout are the only possible values
            event_type = "in" if event_type == "login" else "out"

            event.id = f"{event_type}#{user_id}#{event.created_at}"  # type: ignore

        auth_events_df: DataFrame = authEventsApi.authentication_events_synced_as_df(
            auth_events, sync_db
        )

        return authEventsMap.map_to_udm_system_activities(auth_events_df)

    return _get_authentication_events()



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

from typing import Dict, List, cast
from canvasapi.canvas_object import CanvasObject
from pandas import DataFrame


def _to_dict(canvas_object: CanvasObject):
    """
    Converts a canvas object to a dict containing only the data values

    Parameters
    ----------
    canvas_object: CanvasObject
        a CanvasObject

    Returns
    -------
    Dict
        a Dict containing only the data value
    """
    dictionary = vars(canvas_object).copy()
    del dictionary["_requester"]
    return dictionary


def _to_dicts(canvas_objects: List[CanvasObject]) -> List[Dict]:
    """
    Converts a list of canvas objects to a list of dicts

    Parameters
    ----------
    canvas_objects: List[CanvasObject]
        a list of CanvasObjects

    Returns
    -------
    List[Dict]
        a list of Dicts
    """
    return list(map(_to_dict, canvas_objects))


def to_df(canvas_objects: List) -> DataFrame:
    """
    Converts a list of canvas objects to a DataFrame

    Parameters
    ----------
    canvas_objects: List[CanvasObject]
        a list of CanvasObjects

    Returns
    -------
    DataFrame
        a DataFrame with the data from the canvas objects
    """
    return DataFrame.from_records(
        _to_dicts(cast(List[CanvasObject], canvas_objects))
    ).astype("string")


def remove_duplicates(list: List, identity_property: str) -> List:
    """
    Remove duplicate objects from a list, using a property as the object identity

    Parameters
    ----------
    list: List
        a list of objects
    identity_property: str
        the name of the object property to use as the object identity

    Returns
    -------
    List
        a list of objects with duplicates removed
    """
    unique_dict: Dict = dict()
    for obj in list:
        if getattr(obj, identity_property) not in unique_dict:
            unique_dict[getattr(obj, identity_property)] = obj
    return [*unique_dict.values()]



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.
LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]


class Features:
    Activities = "activities"
    Attendance = "attendance"
    Assignments = "assignments"
    Grades = "grades"  # not implemented yet


VALID_FEATURES = [
    Features.Activities,
    Features.Attendance,
    Features.Assignments,
    Features.Grades,
]



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import pandas as pd
from pandas import DataFrame

from . import constants


def _map_type(lms_type: str):
    return "sign-in" if lms_type == "login" else "sign-out"


def map_to_udm_system_activities(authentication_events: pd.DataFrame) -> pd.DataFrame:
    """
    Maps a DataFrame containing Canvas authentication_events into the Ed-Fi LMS Unified Data
    Model (UDM) format.

    Parameters
    ----------
    authentication_events: DataFrame
        Pandas DataFrame containing authentication_events

    Returns
    -------
    DataFrame
        A LMSSystemActivities-formatted DataFrame

    Notes
    -----
    DataFrame columns are:
        SourceSystemIdentifier: a unique code for the record.
        SourceSystem: the system code or name providing the user data.
        LMSUserSourceSystemIdentifier: the source system identifier for the user.
        ActivityType: either "sign-in" or "sign-out".
        ActivityDateTime: the source system's event timestamp.
        ActivityStatus: will always be "active".
        ParentSourceSystemIdentifier: will always be None.
        ActivityTimeInMinutes: will always be None.
        CreateDate: Created date.
        LastModifiedDate: Last modified date (will always be the same as the CreateDate).
        SourceCreateDate: Date this record was created in the LMS.
    """

    if authentication_events.empty:
        return authentication_events

    assert "CreateDate" in authentication_events.columns
    assert "LastModifiedDate" in authentication_events.columns

    df: DataFrame = authentication_events[
        [
            "id",
            "event_type",
            "created_at",
            "CreateDate",
            "LastModifiedDate",
        ]
    ].copy()

    df["LMSUserSourceSystemIdentifier"] = df["id"].apply(lambda x: x.split("#")[1])

    df.rename(
        columns={
            "id": "SourceSystemIdentifier",
            "event_type": "ActivityType",
            "created_at": "ActivityDateTime",
        },
        inplace=True,
    )

    df["ActivityType"] = df["ActivityType"].apply(_map_type)
    df["SourceSystem"] = constants.SOURCE_SYSTEM
    df["ActivityStatus"] = "active"
    df["ParentSourceSystemIdentifier"] = ""
    df["ActivityTimeInMinutes"] = ""
    df["SourceCreateDate"] = df["ActivityDateTime"]
    df["SourceLastModifiedDate"] = ""
    df["ActivityDateTime"] = pd.to_datetime(df["ActivityDateTime"]).dt.strftime(
        constants.DATE_FORMAT
    )

    return df



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import logging
from typing import List
from pandas import DataFrame
import sqlalchemy

from canvasapi.course import Course
from canvasapi.assignment import Assignment
from edfi_lms_extractor_lib.api.resource_sync import (
    cleanup_after_sync,
    sync_to_db_without_cleanup,
)
from .canvas_helper import to_df
from .api_caller import call_with_retry

ASSIGNMENTS_RESOURCE_NAME = "Assignments"

logger = logging.getLogger(__name__)


def _request_assignments_for_course(course: Course) -> List[Assignment]:
    """
    Fetch Assignments API data for a course

    Parameters
    ----------
    canvas: Canvas
        a Canvas SDK object
    course: Course
        a Canvas Course object

    Returns
    -------
    List[Assignment]
        a list of Assignment API objects
    """
    return call_with_retry(course.get_assignments)


def request_assignments(courses: List[Course]) -> List[Assignment]:
    """
    Fetch Assignments API data for a range of courses and return a list of assignments as Assignment API objects

    Parameters
    ----------
    courses: List[Course]
        a list of Canvas Course objects

    Returns
    -------
    List[Assignment]
        a list of Assignment API objects
    """

    logger.info("Pulling assignment data")
    assignments: List[Assignment] = []
    for course in courses:
        assignments.extend(_request_assignments_for_course(course))

    return assignments


def assignments_synced_as_df(
    assignments: List[Assignment],
    sync_db: sqlalchemy.engine.base.Engine,
) -> DataFrame:
    """
    Fetch Assignments API data for a range of courses and return a Assignments API DataFrame
    with current and previously fetched data

    Parameters
    ----------
    assignments: List[Assignment]
        a list of Canvas Assignments objects
    sync_db: sqlalchemy.engine.base.Engine
        an Engine instance for creating database connections

    Returns
    -------
    DataFrame
        a Assignments API DataFrame with the current and previously fetched data
    """

    assignments_df: DataFrame = _sync_without_cleanup(to_df(assignments), sync_db)
    cleanup_after_sync(ASSIGNMENTS_RESOURCE_NAME, sync_db)

    return assignments_df


def _sync_without_cleanup(
    resource_df: DataFrame, sync_db: sqlalchemy.engine.base.Engine
) -> DataFrame:
    """
    Take fetched API data and sync with database. Creates tables when necessary,
    but ok if temporary tables are there to start. Doesn't delete temporary tables when finished.

    Parameters
    ----------
    resource_df: DataFrame
        a Assignments API DataFrame with the current fetched data which
        will be mutated, adding Hash and CreateDate/LastModifiedDate
    sync_db: sqlalchemy.engine.base.Engine
        an Engine instance for creating database connections

    Returns
    -------
    DataFrame
        a DataFrame with current fetched data and reconciled CreateDate/LastModifiedDate
    """
    return sync_to_db_without_cleanup(
        resource_df=resource_df,
        identity_columns=["id"],
        resource_name=ASSIGNMENTS_RESOURCE_NAME,
        sync_db=sync_db,
    )



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import logging
from typing import List
from pandas import DataFrame
import sqlalchemy

from canvasapi.course import Course
from canvasapi.section import Section
from edfi_lms_extractor_lib.api.resource_sync import (
    cleanup_after_sync,
    sync_to_db_without_cleanup,
)
from .canvas_helper import to_df
from .api_caller import call_with_retry

SECTIONS_RESOURCE_NAME = "Sections"

logger = logging.getLogger(__name__)


def _request_sections_for_course(course: Course) -> List[Section]:
    """
    Fetch Sections API data for a course

    Parameters
    ----------
    canvas: Canvas
        a Canvas SDK object
    course: Course
        a Canvas Course object

    Returns
    -------
    List[Section]
        a list of Section API objects
    """
    return call_with_retry(course.get_sections)


def request_sections(courses: List[Course]) -> List[Section]:
    """
    Fetch Sections API data for a range of courses and return a list of sections as Section API objects

    Parameters
    ----------
    courses: List[Course]
        a list of Canvas Course objects

    Returns
    -------
    List[Section]
        a list of Section API objects
    """

    logger.info("Pulling section data")
    sections: List[Section] = []
    for course in courses:
        sections.extend(_request_sections_for_course(course))

    return sections


def sections_synced_as_df(
    sections: List[Section],
    sync_db: sqlalchemy.engine.base.Engine,
) -> DataFrame:
    """
    Fetch Sections API data for a range of courses and return a Sections API DataFrame
    with current and previously fetched data

    Parameters
    ----------
    sections: List[Section]
        a list of Canvas Sections objects
    sync_db: sqlalchemy.engine.base.Engine
        an Engine instance for creating database connections

    Returns
    -------
    DataFrame
        a Sections API DataFrame with the current and previously fetched data
    """

    sections_df: DataFrame = _sync_without_cleanup(to_df(sections), sync_db)
    cleanup_after_sync(SECTIONS_RESOURCE_NAME, sync_db)

    return sections_df


def _sync_without_cleanup(
    resource_df: DataFrame, sync_db: sqlalchemy.engine.base.Engine
) -> DataFrame:
    """
    Take fetched API data and sync with database. Creates tables when necessary,
    but ok if temporary tables are there to start. Doesn't delete temporary tables when finished.

    Parameters
    ----------
    resource_df: DataFrame
        a Sections API DataFrame with the current fetched data which
        will be mutated, adding Hash and CreateDate/LastModifiedDate
    sync_db: sqlalchemy.engine.base.Engine
        an Engine instance for creating database connections

    Returns
    -------
    DataFrame
        a DataFrame with current fetched data and reconciled CreateDate/LastModifiedDate
    """
    return sync_to_db_without_cleanup(
        resource_df=resource_df,
        identity_columns=["id"],
        resource_name=SECTIONS_RESOURCE_NAME,
        sync_db=sync_db,
    )



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.
import logging
from typing import List
from pandas import DataFrame
import sqlalchemy

from canvasapi import Canvas
from canvasapi.account import Account
from canvasapi.course import Course
from edfi_lms_extractor_lib.api.resource_sync import (
    cleanup_after_sync,
    sync_to_db_without_cleanup,
)
from .api_caller import call_with_retry
from .canvas_helper import to_df

COURSES_RESOURCE_NAME = "Courses"

logger = logging.getLogger(__name__)


def request_courses(canvas: Canvas, start_date: str, end_date: str) -> List[Course]:
    """
    Fetch Course API data for all courses

    Parameters
    ----------
    canvas: Canvas
        a Canvas SDK object

    Returns
    -------
    List[Course]
        a list of Courses API objects
    """
    logger.info("Pulling course data")

    results: List = []

    accounts: List[Account] = call_with_retry(canvas.get_accounts)
    for account in accounts:
        courses: List[Course] = call_with_retry(
            account.get_courses,
            state=["available", "completed"],
            starts_before=end_date,
            ends_after=start_date,
        )
        results.extend(courses)

    return results


def courses_synced_as_df(
    courses: List[Course], sync_db: sqlalchemy.engine.base.Engine
) -> DataFrame:
    """
    Using Course API data, return a Courses API DataFrame

    Parameters
    ----------
    courses: List[Course]
        a list of Canvas Course SDK objects
    sync_db: sqlalchemy.engine.base.Engine
        an Engine instance for creating database connections

    Returns
    -------
    DataFrame
        a Courses API DataFrame with the current and previously fetched data
    """
    courses_df: DataFrame = _sync_without_cleanup(to_df(courses), sync_db)
    cleanup_after_sync(COURSES_RESOURCE_NAME, sync_db)

    return courses_df


def _sync_without_cleanup(
    resource_df: DataFrame, sync_db: sqlalchemy.engine.base.Engine
) -> DataFrame:
    """
    Take fetched API data and sync with database. Creates tables when necessary,
    but ok if temporary tables are there to start. Doesn't delete temporary tables when finished.

    Parameters
    ----------
    resource_df: DataFrame
        a Courses API DataFrame with the current fetched data which
        will be mutated, adding Hash and CreateDate/LastModifiedDate
    sync_db: sqlalchemy.engine.base.Engine
        an Engine instance for creating database connections

    Returns
    -------
    DataFrame
        a DataFrame with current fetched data and reconciled CreateDate/LastModifiedDate
    """
    return sync_to_db_without_cleanup(
        resource_df=resource_df,
        identity_columns=["id"],
        resource_name=COURSES_RESOURCE_NAME,
        sync_db=sync_db,
    )



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.
from typing import Callable
from canvasapi.exceptions import CanvasException
from requests import RequestException
import socket
from opnieuw import retry

MAX_TOTAL_CALLS = 4
RETRY_WINDOW_AFTER_FIRST_CALL_IN_SECONDS = 60


@retry(
    retry_on_exceptions=(
        IOError,
        CanvasException,
        RequestException,
        socket.timeout,
        socket.error,
    ),
    max_calls_total=MAX_TOTAL_CALLS,
    retry_window_after_first_call_in_seconds=RETRY_WINDOW_AFTER_FIRST_CALL_IN_SECONDS,
)
def call_with_retry(canvas_api_function: Callable, **kwargs):
    """
    Invoke a get/list/etc Canvas SDK function that takes kwargs,
    retrying if there are errors.

    Parameters
    ----------
    executable_resource: Callable
        is the get/list/etc Canvas SDK function to call
    kwargs: Dict
        arguments to pass to the Canvas SDK function

    Returns
    -------
    object
        a Canvas SDK response object

    Raises
    ------
    IOError
        if there is an IOError after retrying
    CanvasException
        if there is a CanvasException after retrying
    """

    return canvas_api_function(**kwargs)



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

from datetime import datetime
from pandas import DataFrame

from . import constants


def map_to_udm_users(users_df: DataFrame) -> DataFrame:
    """
    Maps a DataFrame containing Canvas users into the Ed-Fi LMS Unified Data
    Model (UDM) format.

    Parameters
    ----------
    users_df: DataFrame
        Pandas DataFrame containing all Canvas users

    Returns
    -------
    DataFrame
        A LMSUsers-formatted DataFrame

    DataFrame columns are:
        EmailAddress: The primary e-mail address for the user
        LocalUserIdentifier: The user identifier assigned by a school or district
        Name: The full name of the user
        SISUserIdentifier: The user identifier defined in the Student Information System (SIS)
        SourceSystem: The system code or name providing the user data
        SourceSystemIdentifier: A unique number or alphanumeric code assigned to a user by the source system
        CreateDate: datetime at which the record was first retrieved
        LastModifiedDate: datetime when the record was modified, or when first retrieved
        SourceCreateDate: Date this record was created in the LMS
        SourceLastModifiedDate: Date this record was last updated in the LMS
    """

    if users_df.empty:
        return users_df

    df: DataFrame = users_df[
        [
            "id",
            "sis_user_id",
            "created_at",
            "name",
            "email",
            "login_id",
            "CreateDate",
            "LastModifiedDate",
        ]
    ].copy()

    df["SourceSystem"] = constants.SOURCE_SYSTEM

    df.rename(
        columns={
            "id": "SourceSystemIdentifier",
            "sis_user_id": "SISUserIdentifier",
            "login_id": "LocalUserIdentifier",
            "email": "EmailAddress",
            "name": "Name",
            "created_at": "SourceCreateDate",
        },
        inplace=True,
    )

    df["SourceCreateDate"] = df["SourceCreateDate"].apply(
        lambda x: datetime.strftime(
            datetime.strptime(x, "%Y-%m-%dT%H:%M:%S%z"), "%Y/%m/%d %H:%M:%S"
        )
    )
    df["UserRole"] = constants.ROLES.STUDENT

    df["SourceLastModifiedDate"] = ""

    return df



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import logging
from typing import List
import re

from pandas import DataFrame
import sqlalchemy
from canvasapi.authentication_event import AuthenticationEvent
from canvasapi.user import User
from canvasapi.paginated_list import PaginatedList

from .canvas_helper import to_df
from edfi_lms_extractor_lib.api.resource_sync import (
    cleanup_after_sync,
    sync_to_db_without_cleanup,
)
from .api_caller import call_with_retry

AUTH_EVENTS_RESOURCE_NAME = "Authentication_Events"

logger = logging.getLogger(__name__)


def custom_get_new_page(self: PaginatedList):
    response = self._requester.request(
        self._request_method, self._next_url, **self._next_params
    )
    data = response.json()
    self._next_url = None

    next_link = response.links.get("next")
    regex = r"{}(.*)".format(re.escape(self._requester.base_url))

    self._next_url = (
        re.search(regex, next_link["url"]).group(1) if next_link else None  # type: ignore
    )

    self._next_params = {}

    content = []

    if self._root:
        try:
            data = data[self._root]
        except KeyError:
            # TODO: Fix this message to make more sense to an end user.
            raise ValueError("Invalid root value specified.")

    if "audit/authentication/users" in self._first_url:
        data = data["events"]

    for element in data:
        if element is not None:
            element.update(self._extra_attribs)
            content.append(self._content_class(self._requester, element))

    return content


PaginatedList._get_next_page = custom_get_new_page


def _request_events_for_student(
    user: User, start_date: str, end_date: str
) -> List[AuthenticationEvent]:
    def _get_auth_events():
        return user.get_authentication_events(start_time=start_date, end_time=end_date)

    response = call_with_retry(_get_auth_events)
    return response


def request_events(
    users: List[User], start_date: str, end_date: str
) -> List[AuthenticationEvent]:
    """
    Fetch AuthenticationEvent API data for a range of users and return a list of AuthenticationEvent API objects

    Parameters
    ----------
    users: List[User]
        a list of Canvas User objects

    Returns
    -------
    List[AuthenticationEvent]
        a list of AuthenticationEvent API objects
    """

    logger.info("Pulling authentication events data")
    events: List[AuthenticationEvent] = []
    for user in users:
        local_events = _request_events_for_student(user, start_date, end_date)
        events.extend(local_events)

    return events


def authentication_events_synced_as_df(
    auth_events: List[AuthenticationEvent],
    sync_db: sqlalchemy.engine.base.Engine,
) -> DataFrame:
    """
    Fetch AuthenticationEvent API data for a range of courses and return a AuthenticationEvent API DataFrame
    with current and previously fetched data

    Parameters
    ----------
    auth_events: List[AuthenticationEvent]
        a list of Canvas AuthenticationEvent objects
    sync_db: sqlalchemy.engine.base.Engine
        an Engine instance for creating database connections

    Returns
    -------
    DataFrame
        an AuthenticationEvent API DataFrame with the current and previously fetched data
    """
    auth_events_df: DataFrame = _sync_without_cleanup(to_df(auth_events), sync_db)
    cleanup_after_sync(AUTH_EVENTS_RESOURCE_NAME, sync_db)

    return auth_events_df


def _sync_without_cleanup(
    resource_df: DataFrame, sync_db: sqlalchemy.engine.base.Engine
) -> DataFrame:
    """
    Take fetched API data and sync with database. Creates tables when necessary,
    but ok if temporary tables are there to start. Doesn't delete temporary tables when finished.

    Parameters
    ----------
    resource_df: DataFrame
        an AuthenticationEvent API DataFrame with the current fetched data which
        will be mutated, adding Hash and CreateDate/LastModifiedDate
    sync_db: sqlalchemy.engine.base.Engine
        an Engine instance for creating database connections

    Returns
    -------
    DataFrame
        a DataFrame with current fetched data and reconciled CreateDate/LastModifiedDate
    """
    return sync_to_db_without_cleanup(
        resource_df=resource_df,
        identity_columns=["id"],
        resource_name=AUTH_EVENTS_RESOURCE_NAME,
        sync_db=sync_db,
    )



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import logging
from typing import List
from pandas import DataFrame
import sqlalchemy

from canvasapi.assignment import Assignment
from canvasapi.submission import Submission
from edfi_lms_extractor_lib.api.resource_sync import (
    cleanup_after_sync,
    sync_to_db_without_cleanup,
)
from .canvas_helper import to_df
from .api_caller import call_with_retry

SUBMISSIONS_RESOURCE_NAME = "Submissions"

logger = logging.getLogger(__name__)


def _request_submissions_for_assignment(assignment: Assignment) -> List[Submission]:
    """
    Fetch Submissions API data for a assignment

    Parameters
    ----------
    canvas: Canvas
        a Canvas SDK object
    assignment: Assignment
        a Canvas Assignment object

    Returns
    -------
    List[Submission]
        a list of Submission API objects
    """
    return call_with_retry(assignment.get_submissions)


def request_submissions(assignment: Assignment) -> List[Submission]:
    """
    Fetch Submissions API data for a range of assignments and return a list of submissions as Submission API objects

    Parameters
    ----------
    assignment: Assignment
        a Canvas Assignment object

    Returns
    -------
    List[Submission]
        a list of Submission API objects
    """
    return _request_submissions_for_assignment(assignment)


def submissions_synced_as_df(
    submissions: List[Submission],
    sync_db: sqlalchemy.engine.base.Engine,
) -> DataFrame:
    """
    Fetch Submissions API data for a range of assignments and return a Submissions API DataFrame
    with current and previously fetched data

    Parameters
    ----------
    submissions: List[Submission]
        a list of Canvas Submissions objects
    sync_db: sqlalchemy.engine.base.Engine
        an Engine instance for creating database connections

    Returns
    -------
    DataFrame
        a Submissions API DataFrame with the current and previously fetched data
    """
    submissions_df: DataFrame = _sync_without_cleanup(to_df(submissions), sync_db)
    cleanup_after_sync(SUBMISSIONS_RESOURCE_NAME, sync_db)

    return submissions_df


def _sync_without_cleanup(
    resource_df: DataFrame, sync_db: sqlalchemy.engine.base.Engine
) -> DataFrame:
    """
    Take fetched API data and sync with database. Creates tables when necessary,
    but ok if temporary tables are there to start. Doesn't delete temporary tables when finished.

    Parameters
    ----------
    resource_df: DataFrame
        a Submissions API DataFrame with the current fetched data which
        will be mutated, adding Hash and CreateDate/LastModifiedDate
    sync_db: sqlalchemy.engine.base.Engine
        an Engine instance for creating database connections

    Returns
    -------
    DataFrame
        a DataFrame with current fetched data and reconciled CreateDate/LastModifiedDate
    """
    return sync_to_db_without_cleanup(
        resource_df=resource_df,
        identity_columns=["id"],
        resource_name=SUBMISSIONS_RESOURCE_NAME,
        sync_db=sync_db,
    )



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

from datetime import datetime
from pandas import DataFrame

from .constants import SOURCE_SYSTEM


def _get_enrollment_status(status: str) -> str:
    if status == "active":
        return "Active"
    if status == "invited":
        return "Invite pending"
    if status == "inactive":
        return "Archived"
    return "Archived"


def _get_date_formated(date) -> str:
    if not isinstance(date, str) or date == "":
        return ""
    return datetime.strftime(
        datetime.strptime(date, "%Y-%m-%dT%H:%M:%S%z"), "%Y/%m/%d %H:%M:%S"
    )


def map_to_udm_section_associations(enrollments_df: DataFrame) -> DataFrame:
    """
    Maps a DataFrame containing Canvas enrollments into the Ed-Fi LMS Unified Data
    Model (UDM) format.

    Parameters
    ----------
    enrollments_df: DataFrame
        Pandas DataFrame containing all Canvas enrollments

    Returns
    -------
    DataFrame
        A LMSSectionAssociations-formatted DataFrame

    DataFrame columns are:
        SourceSystemIdentifier: A unique number or alphanumeric code assigned to a the section-association by
            the source system
        SourceSystem: The system code or name providing the user data
        EnrollmentStatus: Possible values are Active, Expired, Invite pending, Request Pending, Archived
        LMSUserSourceSystemIdentifier: A unique number or alphanumeric code assigned to a user by the source
            system
        LMSSectionSourceSystemIdentifier: A unique number or alphanumeric code assigned to a section by the
            source system
        CreateDate: Date/time at which the record was first retrieved
        LastModifiedDate: Date/time when the record was modified, or when first retrieved
        SourceCreateDate: Date this record was created in the LMS
        SourceLastModifiedDate: Date this record was last updated in the LMS
    """

    if enrollments_df.empty:
        return enrollments_df

    assert "id" in enrollments_df.columns
    assert "enrollment_state" in enrollments_df.columns
    assert "user_id" in enrollments_df.columns
    assert "course_section_id" in enrollments_df.columns
    assert "created_at" in enrollments_df.columns
    assert "updated_at" in enrollments_df.columns

    enrollments_df = enrollments_df[
        [
            "id",
            "enrollment_state",
            "user_id",
            "course_section_id",
            "created_at",
            "updated_at",
            "CreateDate",
            "LastModifiedDate",
        ]
    ].copy()

    enrollments_df.rename(
        columns={
            "id": "SourceSystemIdentifier",
            "enrollment_state": "EnrollmentStatus",
            "user_id": "LMSUserSourceSystemIdentifier",
            "course_section_id": "LMSSectionSourceSystemIdentifier",
            "created_at": "SourceCreateDate",
            "updated_at": "SourceLastModifiedDate",
        },
        inplace=True,
    )

    enrollments_df["SourceCreateDate"] = enrollments_df["SourceCreateDate"].apply(
        _get_date_formated
    )

    enrollments_df["SourceLastModifiedDate"] = enrollments_df[
        "SourceLastModifiedDate"
    ].apply(_get_date_formated)

    enrollments_df["EnrollmentStatus"] = enrollments_df["EnrollmentStatus"].apply(
        _get_enrollment_status
    )

    enrollments_df["SourceSystem"] = SOURCE_SYSTEM

    return enrollments_df



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

from typing import Dict, cast
from pandas import DataFrame

from . import constants


def map_to_udm_assignments(
    assignments_df: DataFrame, sections_df: DataFrame
) -> Dict[str, DataFrame]:
    """
    Maps a DataFrame containing Canvas assignments into the Ed-Fi LMS Unified Data
    Model (UDM) format.

    Parameters
    ----------
    assignments_df: DataFrame
        Pandas DataFrame containing Canvas assignments
    sections_df: DataFrame
        Pandas DataFrame containing Canvas sections

    Returns
    -------
    DataFrame
        A LMS Assignments-formatted DataFrame

    Notes
    -----
    Assignment DataFrame columns are:
        AssignmentCategory: The category or type of assignment
        AssignmentDescription: The assignment description
        DueDateTime: The date and time the assignment is due
        EndDateTime: The end date and time for the assignment
        LMSSectionSourceSystemIdentifier: A unique numeric identifier assigned
            to the section
        MaxPoints: The maximum number of points a student may receive
        SourceSystem: The system code or name providing the assignment data
        SourceSystemIdentifier: A unique number or alphanumeric code assigned to
            an assignment by the source system
        StartDateTime: The start date and time for the assignment
        Title: The assignment title or name
        SourceCreateDate: Date this record was created in the LMS
        SourceLastModifiedDate: Date this record was last updated in the LMS
        CreateDate: Date this record was created in the extractor
        LastModifiedDate: Date this record was last updated in the extractor
    """

    if assignments_df.empty or sections_df.empty:
        return {}

    assert "id" in assignments_df.columns
    assert "name" in assignments_df.columns
    assert "description" in assignments_df.columns
    assert "created_at" in assignments_df.columns
    assert "updated_at" in assignments_df.columns
    assert "lock_at" in assignments_df.columns
    assert "unlock_at" in assignments_df.columns
    assert "due_at" in assignments_df.columns
    assert "submission_types" in assignments_df.columns
    assert "course_id" in assignments_df.columns
    assert "points_possible" in assignments_df.columns
    assert "CreateDate" in assignments_df.columns
    assert "LastModifiedDate" in assignments_df.columns

    assert "id" in sections_df.columns
    assert "course_id" in sections_df.columns

    assignments_df = assignments_df[
        [
            "id",
            "name",
            "description",
            "created_at",
            "updated_at",
            "lock_at",
            "unlock_at",
            "course_id",
            "points_possible",
            "due_at",
            "submission_types",
            "CreateDate",
            "LastModifiedDate",
        ]
    ].copy()

    sections_df = sections_df[["id", "course_id"]].rename(
        columns={"id": "LMSSectionSourceSystemIdentifier"}
    )
    assignments_df = sections_df.merge(assignments_df, on="course_id").drop(
        columns=["course_id"]
    )

    assignments_df["SourceSystemIdentifier"] = assignments_df[
        ["LMSSectionSourceSystemIdentifier", "id"]
    ].agg("-".join, axis=1)

    assignments_df.drop(columns="id", inplace=True)

    assignments_df.rename(
        columns={
            "name": "Title",
            "description": "AssignmentDescription",
            "created_at": "SourceCreateDate",
            "updated_at": "SourceLastModifiedDate",
            "lock_at": "EndDateTime",
            "unlock_at": "StartDateTime",
            "points_possible": "MaxPoints",
            "due_at": "DueDateTime",
            "submission_types": "SubmissionType",
        },
        inplace=True,
    )

    assignments_df["AssignmentCategory"] = "assignment"
    assignments_df["SourceSystem"] = constants.SOURCE_SYSTEM

    assignments_df["LMSSectionSourceSystemIdentifier"] = assignments_df[
        "LMSSectionSourceSystemIdentifier"
    ].astype("string")

    # group by section id as a Dict of DataFrames
    result: Dict[str, DataFrame] = cast(
        Dict[str, DataFrame],
        dict(tuple(assignments_df.groupby(["LMSSectionSourceSystemIdentifier"]))),
    )

    return result



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import logging
import os

from sqlalchemy import create_engine
import sqlalchemy
from canvasapi import Canvas

logger = logging.getLogger(__name__)


def _is_running_in_notebook() -> bool:
    """
    Determine whether code is running in a Jupyter Notebook

    Returns
    -------
    bool
        True if code is running in a Jupyter Notebook
    """
    main = __import__("__main__", None, None, fromlist=["__file__"])
    return not hasattr(main, "__file__")


def get_sync_db_engine(sync_database_directory: str) -> sqlalchemy.engine.base.Engine:
    """
    Create a SQL Alchemy Engine for a SQLite file

    Returns
    -------
    sqlalchemy.engine.base.Engine
        a SQL Alchemy Engine
    """
    logger.debug(
        "Ensuring database directory at %s", os.path.abspath(sync_database_directory)
    )
    os.makedirs(sync_database_directory, exist_ok=True)
    return create_engine(f"sqlite:///{sync_database_directory}/sync.sqlite")


def get_canvas_api(canvas_base_url: str, canvas_access_token: str) -> Canvas:
    """
    Create new CanvasAPI object for API communication

    Returns
    -------
    Canvas
        a new CanvasAPI object
    """
    return Canvas(canvas_base_url, canvas_access_token)



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import logging
from typing import List
from pandas import DataFrame
import sqlalchemy

from canvasapi.section import Section
from canvasapi.enrollment import Enrollment
from edfi_lms_extractor_lib.api.resource_sync import (
    cleanup_after_sync,
    sync_to_db_without_cleanup,
)
from .canvas_helper import to_df
from .api_caller import call_with_retry

ENROLLMENTS_RESOURCE_NAME = "Enrollments"

logger = logging.getLogger(__name__)


def request_enrollments_for_section(section: Section) -> List[Enrollment]:
    """
    Fetch Enrollments API data for a section

    Parameters
    ----------
    canvas: Canvas
        a Canvas SDK object
    user: User
        a Canvas User object

    Returns
    -------
    List[Enrollment]
        a list of Enrollment API objects
    """
    return call_with_retry(section.get_enrollments)


def enrollments_synced_as_df(
    enrollments: List[Enrollment],
    sync_db: sqlalchemy.engine.base.Engine,
) -> DataFrame:
    """
    Fetch Enrollments API data for a range of students and return a Enrollments API DataFrame
    with current and previously fetched data

    Parameters
    ----------
    enrollments: List[Enrollment]
        a list of Canvas Enrollments objects
    sync_db: sqlalchemy.engine.base.Engine
        an Engine instance for creating database connections

    Returns
    -------
    DataFrame
        a Enrollments API DataFrame with the current and previously fetched data
    """
    enrollments_df: DataFrame = _sync_without_cleanup(to_df(enrollments), sync_db)
    cleanup_after_sync(ENROLLMENTS_RESOURCE_NAME, sync_db)

    return enrollments_df


def _sync_without_cleanup(
    resource_df: DataFrame, sync_db: sqlalchemy.engine.base.Engine
) -> DataFrame:
    """
    Take fetched API data and sync with database. Creates tables when necessary,
    but ok if temporary tables are there to start. Doesn't delete temporary tables when finished.

    Parameters
    ----------
    resource_df: DataFrame
        a Enrollments API DataFrame with the current fetched data which
        will be mutated, adding Hash and CreateDate/LastModifiedDate
    sync_db: sqlalchemy.engine.base.Engine
        an Engine instance for creating database connections

    Returns
    -------
    DataFrame
        a DataFrame with current fetched data and reconciled CreateDate/LastModifiedDate
    """
    return sync_to_db_without_cleanup(
        resource_df=resource_df,
        identity_columns=["id"],
        resource_name=ENROLLMENTS_RESOURCE_NAME,
        sync_db=sync_db,
    )



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

SOURCE_SYSTEM = "Canvas"
INACTIVE = "inactive"
DATE_FORMAT = "%Y-%m-%d %H:%M:%S"


class ROLES:
    STUDENT = "Student"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['edfi_canvas_extractor',
 'edfi_canvas_extractor.api',
 'edfi_canvas_extractor.helpers',
 'edfi_canvas_extractor.mapping']

package_data = \
{'': ['*']}

install_requires = \
['ConfigArgParse>=1.2.3,<2.0.0',
 'SQLAlchemy>=1.3.20,<2.0.0',
 'canvasapi>=2.0.0,<3.0.0',
 'edfi-lms-extractor-lib==1.0.0',
 'edfi-lms-file-utils==1.0.0',
 'errorhandler>=2.0.1,<3.0.0',
 'jupyter>=1.0.0,<2.0.0',
 'opnieuw>=1.1.0,<2.0.0',
 'pandas>=1.1.2,<2.0.0',
 'python-dotenv==0.15.0',
 'requests>=2.24.0,<3.0.0',
 'xxhash>=2.0.0,<3.0.0']

setup_kwargs = {
    'name': 'edfi-canvas-extractor',
    'version': '1.0.0',
    'description': 'Extract tool for retrieving student data from Canvas',
    'long_description': '# Canvas Extractor\n\nThis tool retrieves and writes out to CSV students, active sections,\nassignments, and submissions by querying the Canvas API. For more information on\nthe this tool and its output files, please see the main repository\n[readme](https://github.com/Ed-Fi-Exchange-OSS/LMS-Toolkit).\n\n## Getting Started\n\n1. Download the latest code from [the project\n   homepage](https://github.com/Ed-Fi-Exchange-OSS/LMS-Toolkit) by clicking on\n   the green "CODE" button and choosing an appropriate option. If choosing the\n   Zip option, extract the file contents using your favorite zip tool.\n1. Open a command prompt\\* and change to this file\'s directory (* e.g. cmd.exe,\n   PowerShell, bash).\n1. Ensure you have [Python 3.9+ and Poetry](https://github.com/Ed-Fi-Exchange-OSS/LMS-Toolkit#getting-started).\n1. At a command prompt, install all required dependencies:\n\n   ```bash\n   poetry install\n   ```\n\n1. Optional: make a copy of the `.env.example` file, named simply `.env`, and\n   customize the settings as described in the Configuration section below.\n1. Create an access token by signing into your Canvas instance. Go to Account >\n   Settings, then scroll down to the "Approved Integrations" section and click\n   the "New Access Token" button. If using a `.env` file, then copy the token\n   into this file.\n1. Run the extractor one of two ways:\n   * Execute the extractor with minimum command line arguments:\n\n      ```bash\n      poetry run python edfi_canvas_extractor -b [canvas url] -a [api token]\n          -s [start date range] -e [end date range] -f assignments\n      ```\n\n   * Alternately, run with environment variables or `.env` file:\n\n     ```bash\n     poetry run python edfi_canvas_extractor\n     ```\n\n   * For detailed help, execute `poetry run python edfi_canvas_extractor -h`.\n\n## Configuration\n\nApplication configuration is provided through environment variables or command\nline interface (CLI) arguments. CLI arguments take precedence over environment\nvariables. Environment variables can be set the normal way, or by using a\ndedicated [`.env` file](https://pypi.org/project/python-dotenv/) like\n\n```none\nCANVAS_BASE_URL=[CANVAS_BASE_URL]\nCANVAS_ACCESS_TOKEN=[CANVAS_ACCESS_TOKEN]\nSTART_DATE=[CLASS_START_DATE]\nEND_DATE=[CLASS_END_DATE]\nOUTPUT_DIRECTORY=data\n```\n\nSupported parameters:\n\n| Description | Required | Command Line Argument | Environment Variable |\n| ----------- | -------- | --------------------- | -------------------- |\n| Base Canvas URL | yes | `-b` or `--base-url` | CANVAS_BASE_URL |\n| Canvas API access token | yes | `-a` or `--access-token` | CANVAS_ACCESS_TOKEN |\n| Output Directory | no (default: [working directory]/data) | `-o` or `--output-directory` | OUTPUT_DIRECTORY |\n| Sync database directory | no (default: [working directory]/data) | `-d` or `--sync-database-directory` | SYNC_DATABASE_DIRECTORY |\n| Start date*, yyyy-mm-dd format | yes | `-s` or `--start_date` | START_DATE |\n| End date*, yyyy-mm-dd format | yes | `-e` or `--end_date` | END_DATE |\n| Log level** | no (default: INFO) | `-l` or `--log-level` | LOG_LEVEL |\n| Feature*** | no (default: core, not removable) | `-f` or `--feature` | FEATURE |\n\n\\* _Start Date_ and _End Date_ are used in pulling course and system activity\ndata and would typically span a semester or equivalent school calendar timespan.\n\n\\** Valid values for the optional _log level_:\n\n* DEBUG\n* INFO(default)\n* WARNING\n* ERROR\n* CRITICAL\n\n\\*** When there\'s no specified feature, the extractor will always process Users,\nSections, and Section Associations, which are considered the core feature. Other\nfeatures (can combine two or more):\n\n* assignments (Enables the extraction of assignments and submissions)\n* activities (Enables the extraction of section activities and system\n  activities) - **EXPERIMENTAL**, subject to breaking changes\n* grades (Enables the extraction of grades) - **EXPERIMENTAL**, subject to\n  breaking changes\n\nWhen setting features via `.env` file or through environment variable, combine\nfeatures by using a bracketed comma-separate list, e.g. `FEATURE=[activities,\nattendance, assignments, grades]`. To combine features at the command line,\nsimply list them together: `--feature activities, attendance, assignments,\ngrades]`.\n\n### Output\n\nCSV files in the data(or the specified output) directory with the [LMS Unifying\nData\nModel](https://techdocs.ed-fi.org/display/EDFITOOLS/LMS+Unifying+Data+Model)\nformat.\n\n### Logging and Exit Codes\n\nLog statements are written to the standard output. If you wish to capture log\ndetails, then be sure to redirect the output to a file. For example:\n\n```bash\npoetry run python edfi_canvas_extractor > 2020-12-07-15-43.log\n```\n\nIf any errors occurred during the script run, then there will be a final print\nmessage to the standard error handler as an additional mechanism for calling\nattention to the error: `"A fatal error occurred, please review the log output\nfor more information."`\n\nThe application will exit with status code `1` if there were any log messages at\nthe ERROR or CRITICAL level, otherwise it will exit with status code `0`.\n\n## Developer Operations\n\n1. Style check: `poetry run flake8`\n1. Static typing check: `poetry run mypy .`\n1. Run unit tests: `poetry run pytest`\n1. Run unit tests with code coverage: `poetry run coverage run -m pytest`\n1. View code coverage: `poetry run coverage report`\n\n_Also see\n[build.py](https://github.com/Ed-Fi-Exchange-OSS/LMS-Toolkit/blob/main/docs/build.md)_ for\nuse of the build script.\n\n### Visual Studio Code (Optional)\n\nTo work in Visual Studio Code install the Python Extension.\nThen type `Ctrl-Shift-P`, then choose `Python:Select Interpreter`,\nthen choose the environment that includes `.venv` in the name.\n\n## Legal Information\n\nCopyright (c) 2021 Ed-Fi Alliance, LLC and contributors.\n\nLicensed under the [Apache License, Version\n2.0](https://github.com/Ed-Fi-Exchange-OSS/LMS-Toolkit/blob/main/LICENSE) (the\n"License").\n\nUnless required by applicable law or agreed to in writing, software distributed\nunder the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR\nCONDITIONS OF ANY KIND, either express or implied. See the License for the\nspecific language governing permissions and limitations under the License.\n\nSee [NOTICES](https://github.com/Ed-Fi-Exchange-OSS/LMS-Toolkit/blob/main/NOTICES.md) for\nadditional copyright and license notifications.\n',
    'author': 'Ed-Fi Alliance, LLC, and contributors',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://techdocs.ed-fi.org/display/EDFITOOLS/LMS+Toolkit',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import sys
import logging

from dotenv import load_dotenv
from errorhandler import ErrorHandler

from edfi_canvas_extractor.helpers import arg_parser
from edfi_canvas_extractor.extract_facade import run


def _load_configuration() -> arg_parser.MainArguments:
    load_dotenv()
    return arg_parser.parse_main_arguments(sys.argv[1:])


def _configure_logging(arguments: arg_parser.MainArguments) -> None:
    logging.basicConfig(
        handlers=[
            logging.StreamHandler(sys.stdout),
        ],
        format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
        level=arguments.log_level,
    )

    # Canvas API logging should never be finer-grained than warning, to avoid PII in logs
    canvasapi_logger = logging.getLogger("canvasapi")
    if canvasapi_logger.getEffectiveLevel() < logging.WARN:
        canvasapi_logger.setLevel(logging.WARN)


def _main(arguments):
    error_tracker = ErrorHandler()

    run(arguments)

    if error_tracker.fired:
        print(
            "A fatal error occurred, please review the log output for more information.",
            file=sys.stderr,
        )
        sys.exit(1)
    sys.exit(0)


if __name__ == "__main__":
    arguments = _load_configuration()
    _configure_logging(arguments)
    _main(arguments)



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import pandas as pd
from pandas import DataFrame

from . import constants


def map_to_udm_sections(sections_df: pd.DataFrame) -> pd.DataFrame:
    """
    Maps a DataFrame containing Canvas sections into the Ed-Fi LMS Unified Data
    Model (UDM) format.

    Parameters
    ----------
    sections_df: DataFrame
        Pandas DataFrame containing all Canvas sections

    Returns
    -------
    DataFrame
        A LMSSections-formatted DataFrame

    Notes
    -----
    DataFrame columns are:
        LMSSectionStatus: The section status from the source system
        SISSectionIdentifier: The section identifier defined in the Student Information System
        SectionDescription: The section description
        SourceSystem: The system code or name providing the section data
        SourceSystemIdentifier: A unique number or alphanumeric code assigned to a section
            by the source system
        Term: The enrollment term for the section
        Title: The section title or name
        SourceCreateDate: Date this record was created in the LMS
        SourceLastModifiedDate: Date this record was last updated in the LMS
        CreateDate: Date this record was created in the extractor
        LastModifiedDate: Date this record was last updated in the extractor
    """

    if sections_df.empty:
        return sections_df

    assert "id" in sections_df.columns
    assert "name" in sections_df.columns
    assert "sis_section_id" in sections_df.columns
    assert "CreateDate" in sections_df.columns
    assert "LastModifiedDate" in sections_df.columns

    df: DataFrame = sections_df[
        [
            "id",
            "name",
            "sis_section_id",
            "CreateDate",
            "LastModifiedDate",
        ]
    ].copy()

    df.rename(
        columns={
            "id": "SourceSystemIdentifier",
            "sis_section_id": "SISSectionIdentifier",
            "name": "Title",
        },
        inplace=True,
    )

    df["Term"] = ""
    df["SectionDescription"] = ""
    df["SourceSystem"] = constants.SOURCE_SYSTEM
    df["LMSSectionStatus"] = ""
    df["SourceCreateDate"] = ""
    df["SourceLastModifiedDate"] = ""

    return df



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

from datetime import datetime
from typing import Dict, Tuple, cast
import sys
import logging

from pandas import DataFrame
import sqlalchemy
from canvasapi import Canvas

from edfi_canvas_extractor.config import get_canvas_api, get_sync_db_engine
from edfi_lms_extractor_lib.csv_generation.write import (
    write_section_activities,
    write_users,
    write_sections,
    write_section_associations,
    write_assignments,
    write_grades,
    write_assignment_submissions,
    write_system_activities,
)
from edfi_lms_extractor_lib.helpers.decorators import catch_exceptions

from edfi_canvas_extractor.client_facade import (
    extract_courses,
    extract_grades,
    extract_sections,
    extract_students,
    extract_assignments,
    extract_submissions,
    extract_enrollments,
    extract_system_activities,
)
from edfi_canvas_extractor.api.canvas_helper import to_df
from edfi_canvas_extractor.helpers.arg_parser import MainArguments


logger = logging.getLogger(__name__)


results_store: Dict[str, Tuple] = {}


def _break_execution(failing_extraction: str) -> None:
    logger.critical(
        f"Unable to continue file generation because the load of {failing_extraction} failed. Please review the log for more information."
    )
    sys.exit(1)


@catch_exceptions
def _get_courses(
    arguments: MainArguments, canvas: Canvas, sync_db: sqlalchemy.engine.base.Engine
) -> None:
    logger.info("Extracting Courses from Canvas API")
    (courses, courses_df) = extract_courses(
        canvas, arguments.start_date, arguments.end_date, sync_db
    )
    results_store["courses"] = (courses, courses_df)


@catch_exceptions
def _get_sections(
    arguments: MainArguments, sync_db: sqlalchemy.engine.base.Engine
) -> None:
    logger.info("Extracting Sections from Canvas API")
    (courses, _) = results_store["courses"]
    (sections, udm_sections_df, all_section_ids) = extract_sections(courses, sync_db)
    logger.info("Writing LMS UDM Sections to CSV file")
    write_sections(udm_sections_df, datetime.now(), arguments.output_directory)
    results_store["sections"] = (sections, udm_sections_df, all_section_ids)


@catch_exceptions
def _get_section_activities(
    arguments: MainArguments,
) -> None:
    (_, _, all_section_ids) = results_store["sections"]
    logger.info("Writing empty LMS UDM SectionActivities to CSV files")
    write_section_activities(
        dict(), all_section_ids, datetime.now(), arguments.output_directory
    )


@catch_exceptions
def _get_assignments(
    arguments: MainArguments,
    sync_db: sqlalchemy.engine.base.Engine,
) -> None:
    logger.info("Extracting Assignments from Canvas API")
    (sections, _, all_section_ids) = results_store["sections"]
    (courses, _) = results_store["courses"]
    sections_df = to_df(sections)
    (assignments, udm_assignments_df) = extract_assignments(
        courses, sections_df, sync_db
    )
    logger.info("Writing LMS UDM Assignments to CSV files")
    write_assignments(
        udm_assignments_df, all_section_ids, datetime.now(), arguments.output_directory
    )
    results_store["assignments"] = (assignments, udm_assignments_df)


@catch_exceptions
def _get_students(
    arguments: MainArguments, sync_db: sqlalchemy.engine.base.Engine
) -> None:
    logger.info("Extracting Students from Canvas API")
    (courses, _) = results_store["courses"]
    (students, udm_students_df) = extract_students(courses, sync_db)
    results_store["students"] = (students, udm_students_df)
    logger.info("Writing LMS UDM Users to CSV file")
    write_users(udm_students_df, datetime.now(), arguments.output_directory)


@catch_exceptions
def _get_submissions(
    arguments: MainArguments,
    sync_db: sqlalchemy.engine.base.Engine,
) -> None:
    logger.info("Extracting Submissions from Canvas API")
    (assignments, _) = results_store["assignments"]
    (sections, _, _) = results_store["sections"]
    logger.info("Writing LMS UDM AssignmentSubmissions to CSV files")
    write_assignment_submissions(
        extract_submissions(assignments, sections, sync_db),
        datetime.now(),
        arguments.output_directory,
    )


@catch_exceptions
def _get_enrollments(
    arguments: MainArguments, sync_db: sqlalchemy.engine.base.Engine
) -> None:
    logger.info("Extracting Enrollments from Canvas API")
    (sections, _, all_section_ids) = results_store["sections"]
    (enrollments, udm_enrollments) = extract_enrollments(sections, sync_db)
    logger.info("Writing LMS UDM UserSectionAssociations to CSV files")
    write_section_associations(
        udm_enrollments, all_section_ids, datetime.now(), arguments.output_directory
    )
    results_store["enrollments"] = (enrollments, udm_enrollments)


@catch_exceptions
def _get_grades(arguments: MainArguments) -> None:
    logger.info("Extracting Grades from Canvas API")
    (enrollments, udm_enrollments) = results_store["enrollments"]
    (sections, _, all_section_ids) = results_store["sections"]
    udm_grades: Dict[str, DataFrame] = extract_grades(
        enrollments, cast(Dict[str, DataFrame], udm_enrollments), sections
    )
    logger.info("Writing LMS UDM Grades to CSV files")
    write_grades(
        udm_grades, all_section_ids, datetime.now(), arguments.output_directory
    )


@catch_exceptions
def _get_system_activities(
    arguments: MainArguments, sync_db: sqlalchemy.engine.base.Engine
) -> None:
    logger.info("Extracting System Activities from Canvas API")
    (users, _) = results_store["students"]
    udm_system_activities = extract_system_activities(
        users, arguments.start_date, arguments.end_date, sync_db
    )
    write_system_activities(
        udm_system_activities, datetime.now(), arguments.output_directory
    )


def run(arguments: MainArguments) -> None:
    logger.info("Starting Ed-Fi LMS Canvas Extractor")
    sync_db: sqlalchemy.engine.base.Engine = get_sync_db_engine(
        arguments.sync_database_directory
    )
    succeeded: bool = True

    succeeded = _get_courses(
        arguments, get_canvas_api(arguments.base_url, arguments.access_token), sync_db
    )
    if not succeeded:
        _break_execution("Courses")

    succeeded = _get_sections(arguments, sync_db)
    if not succeeded:
        _break_execution("Sections")

    succeeded = _get_students(arguments, sync_db)
    if not succeeded:
        _break_execution("Students")

    succeeded = _get_enrollments(arguments, sync_db)
    if not succeeded:
        _break_execution("Enrollments")

    if arguments.extract_assignments:
        succeeded = _get_assignments(arguments, sync_db)
        if succeeded:
            _get_submissions(arguments, sync_db)

    if arguments.extract_activities:
        _get_system_activities(arguments, sync_db)
        _get_section_activities(
            arguments,
        )

    if arguments.extract_grades:  # Grades are not supported by all the extractors
        _get_grades(
            arguments,
        )  # Grades don't need sync process because they are part of enrollments

    logger.info("Finishing Ed-Fi LMS Canvas Extractor")



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

from datetime import datetime

import pandas as pd

from . import constants


def _get_date_formated(date) -> str:
    if not isinstance(date, str) or date == "":
        return ""
    return datetime.strftime(
        datetime.strptime(date, "%Y-%m-%dT%H:%M:%S%z"), "%Y/%m/%d %H:%M:%S"
    )


def _get_status(row: pd.Series):
    if row["late"] == 1:
        return "late"
    if row["missing"] == 1:
        return "missing"
    if row["graded_at"] is not None:
        return "graded"
    return "on-time"


def map_to_udm_submissions(submissions_df: pd.DataFrame) -> pd.DataFrame:
    """
    Maps a DataFrame containing Canvas submissions into the Ed-Fi LMS Unified Data
    Model (UDM) format.

    Parameters
    ----------
    submissions_df: DataFrame
        Pandas DataFrame containing Canvas submissions for a section

    Returns
    -------
    DataFrame
        A Submissions-formatted DataFrame .

    Notes
    -----
    DataFrame columns are:
        SourceSystemIdentifier: A unique number or alphanumeric code assigned to a
            submission by the source system
        SourceSystem: The system code or name providing the user data
        SubmissionStatus: The status of the submission
        SubmissionDateTime: Datetime of the submission
        EarnedPoints: Earned points for the submission
        Grade: Grade for the submission
        AssignmentSourceSystemIdentifier: Unique identifier for the assignment
        LMSUserSourceSystemIdentifier: Unique identifier of the LMSUser
        CreateDate: Created date
        LastModifiedDate: Last modified date
        SourceCreateDate: Date this record was created in the LMS
        SourceLastModifiedDate: Date this record was last updated in the LMS
    """

    if submissions_df.empty:
        return submissions_df

    df = submissions_df[
        [
            "id",
            "late",
            "missing",
            "submitted_at",
            "grade",
            "assignment_id",
            "user_id",
            "CreateDate",
            "LastModifiedDate",
            "graded_at",
        ]
    ].copy()

    df["submitted_at"] = df["submitted_at"].apply(_get_date_formated)

    df.rename(
        columns={
            "id": "SourceSystemIdentifier",
            "submitted_at": "SubmissionDateTime",
            "assignment_id": "AssignmentSourceSystemIdentifier",
            "user_id": "LMSUserSourceSystemIdentifier",
            "grade": "Grade",
        },
        inplace=True,
    )

    df["SubmissionDateTime"] = pd.to_datetime(df["SubmissionDateTime"]).dt.strftime(
        constants.DATE_FORMAT
    )
    df["SourceSystem"] = constants.SOURCE_SYSTEM
    df["SubmissionStatus"] = df.apply(_get_status, axis=1)
    df["EarnedPoints"] = None
    df["SourceCreateDate"] = None
    df["SourceLastModifiedDate"] = None

    df.drop(columns=["late", "missing", "graded_at"], inplace=True)

    return df
